<?php

/* COPYRIGHT (C) 2017 Oliver Welter aka Codem@ster
All rights reserved


Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of ISPConfig nor the names of its contributors
      may be used to endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

***************************************************************************************************
SQL for creating the needed tables:

CREATE TABLE `balance` (
  `balance_id` int(10) UNSIGNED NOT NULL,
  `client_id` int(10) UNSIGNED NOT NULL,
  `timestamp_last_update` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP,
  `timestamp_last_booking` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `timestamp_last_loadup` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `timestamp_last_payment` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `timestamp_last_payout` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `timestamp_last_freeze` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `timestamp_last_unfreeze` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `timestamp_last_credit_set` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `timestamp_last_credit_unset` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `timestamp_registration` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `timestamp_disabled` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `timestamp_reenabled` timestamp NOT NULL DEFAULT '0000-00-00 00:00:00',
  `balance` double NOT NULL,
  `credit` double UNSIGNED NOT NULL DEFAULT '0',
  `flag_frozen` tinyint(1) NOT NULL DEFAULT '0',
  `flag_auto_frozen` tinyint(1) NOT NULL DEFAULT '0',
  `flag_have_credit` tinyint(1) NOT NULL DEFAULT '0',
  `flag_disabled` tinyint(1) NOT NULL DEFAULT '0'
) ENGINE=InnoDB DEFAULT CHARSET=utf8;

ALTER TABLE `balance`
  ADD PRIMARY KEY (`balance_id`),
  ADD UNIQUE KEY `client` (`client_id`);

ALTER TABLE `balance`
  MODIFY `balance_id` int(10) UNSIGNED NOT NULL AUTO_INCREMENT;

CREATE TABLE `balance_log` (
  `balance_log_id` int(10) UNSIGNED NOT NULL,
  `balance_id` int(10) UNSIGNED NOT NULL,
  `timestamp_registration` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
  `operation` enum('loadup','payment','payout') NOT NULL,
  `current_balance` double NOT NULL,
  `booking_balance` double NOT NULL,
  `new_balance` double NOT NULL,
  `booking_text` varchar(128) NOT NULL,
  `flag_succeeded` tinyint(1) NOT NULL DEFAULT '0',
  `flag_refund_booking` tinyint(1) NOT NULL DEFAULT '0',
  `refund_balance_id` int(10) UNSIGNED DEFAULT NULL
) ENGINE=InnoDB DEFAULT CHARSET=utf8;

ALTER TABLE `balance_log`
  ADD PRIMARY KEY (`balance_log_id`),
  ADD KEY `balance_id` (`balance_id`);

ALTER TABLE `balance_log`
  MODIFY `balance_log_id` int(10) UNSIGNED NOT NULL AUTO_INCREMENT;

***************************************************************************************************
Also modify ispconfig/interface/web/client/lib/remote.conf.php and add

balance_get,balance_add,balance_set_credit,balance_unset_credit,balance_disable,balance_enable,balance_freeze,balance_unfreeze,balance_get_logs

to

$function_list['client_get_all,...<ADD COMMA SEPARATED STUFF TO THE END>] = 'Client functions';
*/

class remoting_balance extends remoting {
        public function balance_get($session_id, $client_id)
        {
                global $app;

                if(!$this->checkPerm($session_id, 'balance_get')) {
                        throw new SoapFault('permission_denied', 'You do not have the permissions to access this function.');
                        return false;
                }

                $client_id = $app->functions->intval($client_id);
                $sql = "SELECT * FROM balance WHERE client_id  = ?";
                $rec = $app->db->queryOneRecord($sql, $client_id);

                if (!$rec)
                {
                        $this->balance_register($client_id);
                        $rec = $app->db->queryOneRecord($sql, $client_id);
                }

                if (($rec["flag_frozen"] == 0) && ($rec["flag_disabled"] == 0) && (strtotime($rec["timestamp_last_booking"]) < (time() - (86400 * 32))) && ($rec["balance"] < 0))
                {
                        $this->balance_auto_freeze($rec["balance_id"]);
                        $rec = $app->db->queryOneRecord($sql, $client_id);
                }

                return $rec;
        }

        public function balance_get_logs($session_id, $balance_id, $daysback = 1)
        {
                global $app;

                if(!$this->checkPerm($session_id, 'balance_get_logs')) {
                        throw new SoapFault('permission_denied', 'You do not have the permissions to access this function.');
                        return false;
                }

                $balance_id = $app->functions->intval($balance_id);
                $daysback = $app->functions->intval($daysback);

                $limitstart = $app->functions->intval($limitstart);
                $limit = $app->functions->intval($limit);

                $sql = "SELECT * FROM balance_log WHERE balance_id = ?".(($daysback == 0) ? "" : " AND timestampadd(SQL_TSI_DAY, ?, timestamp_registration) > NOW()")." ORDER BY timestamp_registration DESC";

                if ($daysback == 0)
                        $all = $app->db->queryAllRecords($sql, $balance_id);
                else
                        $all = $app->db->queryAllRecords($sql, $balance_id, $daysback);

                return $all;
        }

        public function balance_set($session_id, $balance_id, $balance_value, $booking_text)
        {
                global $app;

                if(!$this->checkPerm($session_id, 'balance_add')) {
                        throw new SoapFault('permission_denied', 'You do not have the permissions to access this function.');
                        return false;
                }

                $balance_id = $app->functions->intval($balance_id);
                $balance = $this->balance_get_by_id($balance_id);

                if (!$balance)
                {
                        throw new SoapFault('permission_denied', 'You do not have the permission to set balance on a non-existent balance-account.');
                        return false;
                }

                $can_book = true;
                $current_balance = $balance["balance"];
                $new_balance = $current_balance + $balance_value;
                $refund_text = "auto refund";

                if ($balance_value < 0)
                {
                        $operation = "payment";

                        if ($balance["flag_have_credit"] == 1)
                        {
                                $total = $balance["balance"] + $balance["credit"];

                                if (($total + $balance_value) < 0)
                                {
                                        $can_book = false;
                                        $refund_text .= " - credit limit exceeded";
                                }
                        }
                        else
                        {
                                if (($balance["balance"] + $balance_value) < 0)
                                {
                                        $can_book = false;
                                        $refund_text .= " - not enough balance";
                                }
                        }

                        if ($balance["flag_frozen"] == 1)
                        {
                                $can_book = false;
                                $refund_text .= " - account frozen";
                        }
                }
                else if ($balance_value == 0)
                {
                        $operation = "payout";

                        $balance_value = $current_balance;
                        $new_balance = 0;

                        if ($balance["flag_frozen"] == 1)
                        {
                                $can_book = false;
                                $refund_text .= " - account frozen";
                        }
                }
                else
                {
                        $operation = "loadup";
                }

                if ($balance["flag_disabled"] == 1)
                {
                        $can_book = false;
                        $refund_text .= " - account disabled";
                }

                $balance_log_id = $this->balance_log($balance_id, $operation, $current_balance, $balance_value, $new_balance, $booking_text);

                $sql = "UPDATE balance SET balance = balance + ?, timestamp_last_booking = NOW(), timestamp_last_".$operation." = NOW() WHERE balance_id = ?";

                if ($can_book)
                {
                        if (!$app->db->query($sql, (double)$balance_value, $balance_id))
                        {
                                if (substr($balance_value, 0, 1) == "-")
                                        $nbal = (double)substr($balance_value, 1);
                                else
                                        $nbal = (double)"-".$balance_value;

                                $refund_log_id = $this->balance_log($balance_id, $operation, $new_balance, $nbal, $current_balance, $refund_text, $balance_log_id);
                                $this->balance_log_flag_succeeded($refund_log_id);

                                return false;
                        }

                        $this->balance_log_flag_succeeded($balance_log_id);

                        return true;
                }
                else
                {
                        if (substr($balance_value, 0, 1) == "-")
                                $nbal = (double)substr($balance_value, 1);
                        else
                                $nbal = (double)"-".$balance_value;

                        $refund_log_id = $this->balance_log($balance_id, $operation, $new_balance, $nbal, $current_balance, $refund_text, $balance_log_id);

                        $this->balance_log_flag_succeeded($refund_log_id);

                        return false;
                }
        }

        public function balance_set_credit($session_id, $balance_id, $credit)
        {
                global $app;

                if(!$this->checkPerm($session_id, 'balance_set_credit')) {
                        throw new SoapFault('permission_denied', 'You do not have the permissions to access this function.');
                        return false;
                }

                if ($credit < 0)
                {
                        throw new SoapFault('permission_denied', 'You do not have the permissions to give negative credits.');
                        return false;
                }

                $balance_id = $app->functions->intval($balance_id);
                $sql = "UPDATE balance SET flag_have_credit = 1, timestamp_last_credit_set = NOW(), credit = ? WHERE balance_id = ?";

                return $app->db->query($sql, (double)$credit, $balance_id);
        }

        public function balance_unset_credit($session_id, $balance_id)
        {
                global $app;

                if(!$this->checkPerm($session_id, 'balance_unset_credit')) {
                        throw new SoapFault('permission_denied', 'You do not have the permissions to access this function.');
                        return false;
                }

                $balance_id = $app->functions->intval($balance_id);
                $sql = "UPDATE balance SET flag_have_credit = 0, timestamp_last_credit_unset = NOW() WHERE balance_id = ?";

                return $app->db->query($sql, $balance_id);
        }

        public function balance_disable($session_id, $balance_id)
        {
                global $app;

                if(!$this->checkPerm($session_id, 'balance_disable')) {
                        throw new SoapFault('permission_denied', 'You do not have the permissions to access this function.');
                        return false;
                }

                $balance_id = $app->functions->intval($balance_id);
                $sql = "UPDATE balance SET flag_disabled = 1, timestamp_disabled = NOW() WHERE balance_id = ?";

                return $app->db->query($sql, $balance_id);
        }

        public function balance_enable($session_id, $balance_id)
        {
                global $app;

                if(!$this->checkPerm($session_id, 'balance_enable')) {
                        throw new SoapFault('permission_denied', 'You do not have the permissions to access this function.');
                        return false;
                }

                $balance_id = $app->functions->intval($balance_id);
                $sql = "UPDATE balance SET flag_disabled = 0, timestamp_reenabled = NOW() WHERE balance_id = ?";

                return $app->db->query($sql, $balance_id);
        }

        public function balance_freeze($session_id, $balance_id)
        {
                global $app;

                if(!$this->checkPerm($session_id, 'balance_freeze')) {
                        throw new SoapFault('permission_denied', 'You do not have the permissions to access this function.');
                        return false;
                }

                $balance_id = $app->functions->intval($balance_id);
                $sql = "UPDATE balance SET flag_frozen = 1, timestamp_last_freeze = NOW() WHERE balance_id = ?";

                return $app->db->query($sql, $balance_id);
        }

        public function balance_unfreeze($session_id, $balance_id)
        {
                global $app;

                if(!$this->checkPerm($session_id, 'balance_unfreeze')) {
                        throw new SoapFault('permission_denied', 'You do not have the permissions to access this function.');
                        return false;
                }

                $balance_id = $app->functions->intval($balance_id);
                $sql = "UPDATE balance SET flag_frozen = 0, flag_auto_frozen = 0, timestamp_last_unfreeze = NOW() WHERE balance_id = ?";

                return $app->db->query($sql, $balance_id);
        }

        private function balance_get_by_id($balance_id)
        {
                global $app;

                $balance_id = $app->functions->intval($balance_id);
                $sql = "SELECT * FROM balance WHERE balance_id = ?";

                $balance = $app->db->queryOneRecord($sql, $balance_id);

                if (($balance["flag_frozen"] == 0) && ($balance["flag_disabled"] == 0) && (strtotime($balance["timestamp_last_booking"]) < (time() - (86400 * 32))) && ($balance["balance"] < 0))
                {
                        $this->balance_auto_freeze($balance_id);
                        $balance = $app->db->queryOneRecord($sql, $balance_id);
                }

                return $balance;
        }

        private function balance_log_flag_succeeded($balance_log_id)
        {
                global $app;

                $balance_log_id = $app->functions->intval($balance_log_id);
                $sql = "UPDATE balance_log SET flag_succeeded = 1 WHERE balance_log_id = ?";

                return $app->db->query($sql, $balance_log_id);
        }

        private function balance_log($balance_id, $operation, $current_balance, $booking_balance, $new_balance, $booking_text, $refund_balance_id = null)
        {
                global $app;

                $balance_id = $app->functions->intval($balance_id);

                if ($refund_balance_id === null)
                {
                        $sql = "INSERT INTO balance_log (balance_id,operation,current_balance,booking_balance,new_balance,booking_text) VALUES (?,?,?,?,?,?)";
                        $app->db->query($sql, $balance_id, (string)$operation, (double)$current_balance, (double)$booking_balance, (double)$new_balance, (string)$booking_text);
                }
                else
                {
                        $refund_balance_id = $app->functions->intval($refund_balance_id);
                        $sql = "INSERT INTO balance_log (balance_id,operation,current_balance,booking_balance,new_balance,booking_text,flag_refund_booking,refund_balance_id) VALUES (?,?,?,?,?,?,1,?)";
                        $app->db->query($sql, $balance_id, (string)$operation, (double)$current_balance, (double)$booking_balance, (double)$new_balance, (string)$booking_text, $refund_balance_id);
                }

                return $app->db->insertID();
        }

        private function balance_auto_freeze($balance_id)
        {
                global $app;

                $balance_id = $app->functions->intval($balance_id);
                $sql = "UPDATE balance SET flag_frozen = 1, flag_auto_frozen = 1, timestamp_last_freeze = NOW() WHERE balance_id = ?";

                return $app->db->query($sql, $balance_id);
        }

        private function balance_register($client_id)
        {
                global $app;

                $client_id = $app->functions->intval($client_id);
                $sql = "INSERT INTO balance (client_id,timestamp_registration,balance) VALUES (?,NOW(),0)";
                $app->db->query($sql, $client_id);
        }
}

?>
